"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const midi_1 = require("@tonejs/midi");
const helpers_1 = require("./helpers");
const notes_processor_1 = __importDefault(require("./notes-processor"));
function midiHasMultipleTempos(midi) {
    return midi.header.tempos.length !== 1;
}
function midiHasMultipleTimeSignatures(midi) {
    return midi.header.timeSignatures.length !== 1;
}
function midiHasNonFourBasedTimeSignature(midi) {
    return midi.header.timeSignatures[0].timeSignature[1] !== 4;
}
function isPolyphonicMidiFile(midi) {
    function noteSequenceIsPolyphonic(notes) {
        const timeDiffs = notes_processor_1.default.getTimeDifferenceArray(notes);
        return timeDiffs.filter((num) => num < 0).length > 0;
    }
    return (0, helpers_1.getValidTracks)(midi)
        .map((track) => track.notes)
        .map(noteSequenceIsPolyphonic)
        .some((b) => !!b);
}
class SimpleMidi extends midi_1.Midi {
    constructor(buffer, bpm, timeSignature) {
        super(buffer);
        this.simpleBpm = bpm;
        this.simpleTimeSignature = timeSignature;
    }
    static validateMidi(midi) {
        if (midiHasMultipleTempos(midi)) {
            throw new Error("SimpleMidi must have only 1 tempo!");
        }
        if (midiHasMultipleTimeSignatures(midi)) {
            throw new Error("SimpleMidi must have only 1 time signature!");
        }
        //
        if (midiHasNonFourBasedTimeSignature(midi)) {
            throw new Error("SimpleMidi time signature must have bottom number of 4!");
        }
        if (isPolyphonicMidiFile(midi)) {
            throw new Error("Midi file contains polyphony. This isn't supported yet. Please make sure your midi file contains no tracks that have overlapping notes.");
        }
    }
    static fromBuffer(buffer) {
        try {
            const midi = new midi_1.Midi(buffer);
            SimpleMidi.validateMidi(midi);
            const bpm = midi.header.tempos[0].bpm;
            const timeSignature = midi.header.timeSignatures[0].timeSignature;
            return new SimpleMidi(buffer, bpm, timeSignature);
        }
        catch (e) {
            console.info("Buffer isn't valid Midi or SimpleMidi");
            console.info(e);
            return null;
        }
    }
    static fromBase64(str) {
        return SimpleMidi.fromBuffer(Buffer.from(str, "base64"));
    }
}
exports.default = SimpleMidi;
